******************************************************************
* COPYRIGHT (C) 1986 by Donald Krantz and James Stanley
* - Note: This is a real, live, actual, registered copyright,
*   and should be treated as such. This source code is from
*   the book "68000 Assembly Language", Krantz and Stanley,
*   Addison-Wesley Publishing Company, Reading, MA, 1986.
* 
*   Permission granted by the authors for non-commercial use
*   in programs released to the public domain, as long as this
*   copyright notice remains attached and visible.
*
*****************************************************************
* OKIDATA graphics printer interface 

#bit.h

	xref	image
	xdef	dump

*****************************************************************
* Column offsets into graphics image area
row1	equ	hor_w*2
row2	equ	hor_w*4
row3	equ	hor_w*6
row4	equ	hor_w*8
row5	equ	hor_w*10
row6	equ	hor_w*12
*****************************************************************
* Prints the graphic image on the ol' Okidata ML-83A
dump:
	movem.l	d0-d7/a0,-(a7)	* Save all registers
	bsr	setup_lst	* set up printer		
	move.w	#((vert/7)),d0  D0 will be row loop counter
******* Outer loop: prints 7 rows per pass thru this loop *******
outer:
	swap	d0		* keep counter in hi word
	move.l	#image,a0	* a0 will be image base
	move.l	d0,d7		* bring row counter into d7	
	swap	d7		* D7.W is row counter
	sub.w	#((vert/7)),d7  adjust for reverse direction
	neg.w	d7		* take absolute value
	mulu	#(hor_w*14),d7	* get offset to this row
	add.l	d7,a0		* A0 points at correct top row
	move.w	#((2*hor_w)-1),d1 d1 will be word loop counter
******* Middle loop: Picks up 7 vertically aligned bytes per pass
middle:
	swap	d1		* keep counter in hi word
	move.w	#7,d7		* d7 will be bit counter
	move.b	(a0),d0		* get top (row 1 ) bits
	move.b	row1(a0),d1	* Get row 2 bits
	move.b	row2(a0),d2	* Get row 3 bits
	move.b	row3(a0),d3	* Get row 4 bits
	move.b	row4(a0),d4	* Get row 5 bits
	move.b	row5(a0),d5	* Get row 6 bits
	move.b	row6(a0),d6	* Get bottom (row 7) bits
******* Inner loop: Cranks out one graphic byte per pass ********
inner:
	swap	d7		* keep counter in high word
	roxr.b	d6		* position bottom (row 7) bit
	roxl.b	d7		* pull extend flag into d7
 	roxr.b	d5		* position row 6 bit in ext flag
	roxl.b	d7		* pull extend flag into d7
	roxr.b	d4		* position row 5 bit in ext flag
	roxl.b	d7		* pull extend flag into d7
	roxr.b	d3		* position row 4 bit in ext flag
	roxl.b	d7		* pull extend flag into d7
	roxr.b	d2		* position row 3 bit in ext flag
	roxl.b	d7		* pull extend flag into d7
	roxr.b	d1		* position row 2 bit in ext flag
	roxl.b	d7		* pull extend flag into d7
	roxr.b	d0		* position top (row 1) bit
	roxl.b	d7		* pull extend flag into d7
	bsr	putd7		* output byte in d7
	swap	d7		* bring bit cntr into position
	dbra	d7,inner	* loop for the next col
******* end of inner loop ***************************************
	addq.l	#1,a0		* bump word pointer
	swap	d1		* bring word cntr into position
	dbra	d1,middle	* loop for the next word
******* end of middle loop **************************************
	bsr	gcrlf		* graphics return/line feed
	swap	d0		* bring row cntr into position
	dbra	d0,outer	* loop for next rows
******* end of outer loop ***************************************
	bsr	unset_lst	* reset printer
	movem.l	(a7)+,d0-d7/a0	* restore all registers
	rts
*****************************************************************
* Puts printer into graphics mode
setup_lst:
	move.w	#3,d1		* OKI enter graphics command
	bsr	lst_out		* printer output
	rts
*****************************************************************
* Removes printer from graphics mode
unset_lst:
	move.w	#3,d1		* OKI graphics escape
	bsr	lst_out		* printer output
	move.w	#2,d1		* OKI exit graphics command
	bsr	lst_out		* printer output
	rts
*****************************************************************
* Outputs byte in D7 as graphics byte - traps escape byte
putd7:
	movem.l	d0/d1,-(a7)	* save registers
	move.w	d7,d1		* get output byte lined up
	and.w	#$007F,d1	* off with its head!
	cmp.b	#3,d1		* test for graphics escape char
	bne	skip_put	* jump if not escape char
	move.w	#3,d1		* escape cancel prefix
	bsr	lst_out		* output to printer
	move.w	#3,d1		* reload esacpe char
skip_put:
	bsr	lst_out		* output to printer
	movem.l	(a7)+,d0/d1	* restore registers
	rts
*****************************************************************
* Outputs a return/linefeed sequence in graphics mode
gcrlf:
	movem.l	d0/d1,-(a7)	* Save registers
	move.w	#3,d1		* graphics escape code
	bsr	lst_out		* output to printer
	move.w	#14,d1		* grpahics return/line feed
	bsr	lst_out		* output to printer
	movem.l	(a7)+,d0/d1	* restore registers
	rts
*****************************************************************
* Outputs byte in D1 to CP/M LST: device
* destroys D0 and D1
lst_out:
	move.w	#5,d0		* CP/M LST: device output code
	trap	#2		* Call CP/M BDOS
	rts
	
	end
